/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * GtkDisplay: Emulate a simple 7 segment digit
 * Copyright (C) 1999 Uwe Steinmann
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
//#include <config.h>
#include <stdio.h>
#include "gtkled.h"
#include "lcdfont.h"

#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

static void gtk_led_class_init	 (GtkLedClass	 *klass);
static void gtk_led_init	 (GtkLed	 *led);
static void gtk_led_destroy	 (GtkObject	 *object);
#if 0
static void gtk_led_size_request (GtkWidget	 *widget,
				  GtkRequisition *requisition);
#endif
static void gtk_led_size_allocate (GtkWidget	 *widget,
				  GtkAllocation *allocation);
static void gtk_led_realize      (GtkWidget      *widget);
static void gtk_led_draw (GtkWidget         *widget, GdkRectangle      *area);


static GtkMiscClass *parent_class = NULL;

enum {
  LED_COLOR_ON,
  LED_COLOR_OFF
};

GtkType
gtk_led_get_type ()
{
  static GtkType led_type = 0;
  
  if (!led_type)
  {
    GtkTypeInfo led_info =
    {
      "GtkLed",
      sizeof (GtkLed),
      sizeof (GtkLedClass),
      (GtkClassInitFunc) gtk_led_class_init,
      (GtkObjectInitFunc) gtk_led_init,
      /* reserved */ NULL,
      /* reserved */ NULL,
      (GtkClassInitFunc) NULL,
    };
    
    led_type = gtk_type_unique (gtk_misc_get_type (), &led_info);
  }
  
  return led_type;
}

void
gtk_led_class_init (GtkLedClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  
  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;
  
  parent_class = gtk_type_class (gtk_misc_get_type ());
  
  object_class->destroy = gtk_led_destroy;
  
//  widget_class->size_request  = gtk_led_size_request;
  widget_class->size_allocate = gtk_led_size_allocate;
  widget_class->realize       = gtk_led_realize;
  widget_class->draw          = gtk_led_draw;
}

void
gtk_led_init (GtkLed *led)
{
  GtkMisc *misc;
  
  misc = GTK_MISC (led);

  GTK_WIDGET_SET_FLAGS (led, GTK_NO_WINDOW);
  
  led->value.int_value   = 0;
  led->gc                = NULL;
  led->ledtype           = DigitAlphaNumeric;
	if(NULL == (led->grid = (char *) g_malloc(ROWS * COLS * sizeof(char)))) {
		fprintf(stderr, "Could not allocate memory for led grid\n");
	}
}

GtkWidget*
gtk_led_new ()
{
  GtkLed *led;
  
  led = gtk_type_new (gtk_led_get_type ());

  return GTK_WIDGET (led);
}

void
gtk_led_set_colors (GtkLed *led, GdkColor *active, GdkColor *inactive)
{
	GdkColormap  *cmap;

	g_return_if_fail (led != NULL);
	g_return_if_fail (GTK_IS_LED (led));

	led->fg[LED_COLOR_OFF] = *(inactive);
	led->fg[LED_COLOR_ON] = *(active);

	cmap = gtk_widget_get_colormap (GTK_WIDGET(led));
      
	if (!(&(led->fg[LED_COLOR_ON])))
		gdk_color_parse ("#00F100", &(led->fg[LED_COLOR_ON]));
	gdk_color_alloc (cmap, &(led->fg[LED_COLOR_ON]));
	if (!(&(led->fg[LED_COLOR_OFF])))
		gdk_color_parse ("#008C00", &(led->fg[LED_COLOR_OFF]));
	gdk_color_alloc (cmap, &(led->fg[LED_COLOR_OFF]));
}

void
gtk_led_set_state (GtkLed	*led,
		   GtkStateType widget_state,
		   guint	value)
{
  g_return_if_fail (led != NULL);
  g_return_if_fail (GTK_IS_LED (led));

  gtk_widget_set_state (GTK_WIDGET (led), widget_state);
  gtk_led_set_value (led, value);
}

void
gtk_led_set_value (GtkLed	     *led,
		guint     value)
{
	struct char_def *chr;
	char *grid;
	int i, j;

  g_return_if_fail (led != NULL);
  g_return_if_fail (GTK_IS_LED (led));

	if(led->ledtype == DigitAlphaNumeric) {
  	led->value.char_value = (char) value;
		grid = led->grid;
		chr = &lcdfont[value];
		for( i = 0; i < ROWS; i++ )
			for( j = 0; j < COLS; j++ )
				if(chr->data[i] & (0x10 >> j))
					grid[i*COLS+j] = 1;
				else
					grid[i*COLS+j] = 0;
	} else {
  	led->value.int_value = value;
	}
  gtk_widget_draw (GTK_WIDGET (led), NULL);
}

guint
gtk_led_get_value(GtkLed  *led)
{
  g_return_val_if_fail (led != NULL, -1);
  g_return_val_if_fail (GTK_IS_LED (led), -1);
  
  return led->value.int_value;
}

void
gtk_led_set_ledtype (GtkLed	     *led,
		LEDTYPE     value)
{
  g_return_if_fail (led != NULL);
  g_return_if_fail (GTK_IS_LED (led));

  led->ledtype = value;
}

LEDTYPE
gtk_led_get_ledtype(GtkLed  *led)
{
  g_return_val_if_fail (led != NULL, -1);
  g_return_val_if_fail (GTK_IS_LED (led), -1);
  
  return led->ledtype;
}

static void
gtk_led_destroy (GtkObject *object)
{
  GtkLed *led;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_LED (object));
 
  led = GTK_LED (object);
	if(led->grid)
		g_free(led->grid);
  
  if (GTK_WIDGET (object)->parent &&
      GTK_WIDGET_MAPPED (object))
    gtk_widget_unmap (GTK_WIDGET (object));
  
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_led_size_allocate (GtkWidget	     *widget,
		      GtkAllocation *allocation)
{
  GtkLed *led;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_LED (widget));
  g_return_if_fail (allocation != NULL);
  
//	fprintf(stderr, "led: size_allocate called: %d, %d\n", allocation->width, allocation->height);
  led = GTK_LED (widget);
  
	widget->allocation = *allocation;
}

#if 0
static void
gtk_led_size_request (GtkWidget	     *widget,
		      GtkRequisition *requisition)
{
  GtkLed *led;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_LED (widget));
  g_return_if_fail (requisition != NULL);
  
  led = GTK_LED (widget);
  
  requisition->width = LED_WIDTH;
  requisition->height = LED_HEIGHT;
}
#endif

static void 
gtk_led_realize (GtkWidget *widget)
{
  GtkLed       *led;
  GdkColormap  *cmap;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_LED (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  led = GTK_LED (widget);
  widget->window = gtk_widget_get_parent_window (widget);
  gdk_window_ref (widget->window);
  widget->style = gtk_style_attach (widget->style, widget->window);
  if (!led->gc)
    {
      cmap = gtk_widget_get_colormap (widget);
      
      if (!(&(led->fg[LED_COLOR_ON])))
				gdk_color_parse ("#00F100", &(led->fg[LED_COLOR_ON]));
      gdk_color_alloc (cmap, &(led->fg[LED_COLOR_ON]));
      if (!(&(led->fg[LED_COLOR_OFF])))
					gdk_color_parse ("#008C00", &(led->fg[LED_COLOR_OFF]));
      gdk_color_alloc (cmap, &(led->fg[LED_COLOR_OFF]));
      led->gc = gdk_gc_new (widget->window);
      gdk_gc_copy (led->gc, widget->style->white_gc);
    }
	

}

/*  0000
   3    4
   3    4
    1111
   5    6
   5    6
    2222
*/
static gint
led_lit (int pos, int val) {
	static int lookup[12]= {125,80,55,87,90,79,111,81,127,95,2,0};
	if(val >= 48) {
		val -= 48;
		if(val > 11)
			val = 11;
	} else if (val == 45)
	  val = 10;
	else if (val == 32)
		val = 11;
	else if(val < 48)
		val = 11;
	// lookup[10] is a minus sign
	// lookup[11] is all segments off
	return ((1<<pos)&lookup[val]);
}

static void
gtk_led_draw(GtkWidget *widget, GdkRectangle *area) {
  GtkLed   *led;
  GtkMisc  *misc;
  GdkColor *win_bg;
  gint     i;
  gint     w, val, start_x, start_y;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_LED (widget));
  g_return_if_fail (area != NULL);
  
  led = GTK_LED (widget);
  misc = GTK_MISC (widget);

	if (GTK_WIDGET_DRAWABLE (widget)) {
		if ((widget->allocation.width >= widget->requisition.width) &&
		    (widget->allocation.height >= widget->requisition.height)) {
			guint x, y;
			win_bg = &(led->fg[LED_COLOR_OFF]);
			gdk_gc_set_foreground (led->gc, win_bg);
			x = widget->allocation.x;
			y = widget->allocation.y;

//fprintf(stderr, "gtkled: height = %d, width = %d\n", widget->allocation.height, widget->allocation.width);
			if(widget->allocation.height < 5)
				return;
			if(widget->allocation.width < 5)
				return;

			gdk_draw_rectangle (widget->window,
			  led->gc,
			  TRUE,
			  x + 0, y + 0,
			  widget->allocation.width - 1,
			  widget->allocation.height - 1);

			if(led->ledtype == DigitSevenSegment) {
				val = led->value.int_value;
				start_x = x+2;
				start_y = y+2;
				x = widget->allocation.width - 5;
				y = widget->allocation.height/2-4;
				w = MIN(x,y)/6;
				if (w<3) w=2;
				if (!(w%2)) w++;
	
				for (i=0;i<w;i++) {
					if (led_lit(0,val)) gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_ON]));
					else gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_OFF]));
					gdk_draw_line (widget->window, led->gc,
					           start_x+w/2+i, start_y+i, start_x+x-w/2-i-1, start_y+i);
					/* Middle segment, Bit 1 */
					if (led_lit(1,val)) gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_ON]));
					else gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_OFF]));
					if (!(i%2)) gdk_draw_line (widget->window, led->gc,
					                       start_x+w/2+(i/2)+1,   start_y+y-(i/2)+1,
					                       start_x+x-w/2-(i/2)-2, start_y+y-(i/2)+1);
					else     gdk_draw_line (widget->window, led->gc,
					                    start_x+w/2+(i/2)+2,       start_y+y+(i/2)+2,
					                    start_x+x-w/2-((i/2)+3), start_y+y+(i/2)+2);
					if (led_lit(2,val)) gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_ON]));
					else gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_OFF]));
					gdk_draw_line (widget->window, led->gc,
					           start_x+w/2+i,   start_y+2*y-i+2,
					           start_x+x-w/2-i-1, start_y+2*y-i+2);
					if (led_lit(3,val)) gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_ON]));
					else gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_OFF]));
					gdk_draw_line (widget->window, led->gc,
					           start_x+i, start_y+w/2+i+1, start_x+i, start_y+y-i-w/2+1);
					if (led_lit(4,val)) gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_ON]));
					else gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_OFF]));
					gdk_draw_line (widget->window, led->gc,
					           start_x+x-i-1, start_y+w/2+i+1, start_x+x-i-1, start_y+y-i-w/2+1);
					if (led_lit(5,val)) gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_ON]));
					else gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_OFF]));
					gdk_draw_line (widget->window, led->gc,
					           start_x+i, start_y+w/2+i+y+1, start_x+i, start_y+2*y-i-w/2+1);
					if (led_lit(6,val)) gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_ON]));
					else gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_OFF]));
					gdk_draw_line (widget->window, led->gc,
					           start_x+x-i-1, start_y+w/2+i+y+1, start_x+x-1-i, start_y+2*y-i-w/2+1);
				}
			} else { // not SEVEN_SEGMENT
				float boxwidth, boxheight;
				int i, j;
				int iBorder = 2;

				boxwidth = (float) (((widget->allocation.width-2)-2*iBorder) /
				                    (float) COLS);
				boxheight = (float) (((widget->allocation.height-1)-2*iBorder) /
				                      (float) ROWS);
				gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_ON]));
				for( i = 0; i < ROWS; i++ )
					for( j = 0; j < COLS; j++ ) {
						if(led->grid[i*COLS+j]) {
							gdk_draw_rectangle(widget->window,
							                   led->gc, TRUE,
							                   (int) (x+j*boxwidth+iBorder),
							                   (int) (y+i*boxheight+iBorder),
							                   (int) (boxwidth+1),
							                   (int) (boxheight+1));
						}
					}

				if((boxwidth < 2.0) || (boxheight < 2.0)) {
				} else {
					gdk_gc_set_foreground(led->gc, &(led->fg[LED_COLOR_OFF]));
					for( i = 0; i <= ROWS; i++ )
						gdk_draw_line (widget->window, led->gc,
					               	x+iBorder,
												 	(int) (y + i * boxheight + iBorder),
												 	(int) (x + COLS * boxwidth + iBorder),
												 	(int) (y + i * boxheight + iBorder) );
					for( i = 0; i <= COLS; i++ )
						gdk_draw_line (widget->window, led->gc,
					               	(int) (x + i * boxwidth + iBorder),
					               	y + iBorder,
												 	(int) (x + i * boxwidth + iBorder),
												 	(int) (y + ROWS * boxheight + iBorder) );

				}
			}
		}
	}
}
/* EOF */

