#include "imap.h"
#include "util.h"

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#define IMAP_DEFAULT_PORT 143

typedef void (*UntaggedCallback)(ImapContext *, char *, char *);

void handle_ignore(ImapContext *c, char *resp, char *args) {
  return;
}

void handle_status(ImapContext *c, char *resp, char *args) {
  char *p;

  if(strcasecmp(resp, "STATUS"))
    return;

  if(!args)
    goto error;

  /* Kill the final close paren */
  p = args + strlen(args) - 1;
  *p = '\0';

  /* And then find the first one */
  p = strchr(args, '(');
  if(!p)
    goto error;

  p = strchr(p, ' ');
  if(!p)
    goto error;

  p++;
  c->recent = atoi(p);
  return;

error:
  c->recent = 0;
}

static void imap_error(ImapContext *c, const char *fmt, ...) {
  va_list ap;

  c->error = 1;

  va_start(ap, fmt);
  vsnprintf(c->errstr, IMAP_BUFSIZE, fmt, ap);
  va_end(ap);
}

static void imap_close(ImapContext *c) {
  if(c->out) {
    fprintf(c->out, ". LOGOUT\r\n");
    fclose(c->out);
  }
  if(c->in)
    fclose(c->in);
}

static ImapContext *imap_new() {
  ImapContext *c;

  c = malloc(sizeof(ImapContext));
  if(!c)
    return NULL;

  c->in = c->out = NULL;
  c->error = 0;
  c->recent = 0;
  memset(c->buf, 0, IMAP_BUFSIZE);
  memset(c->errstr, 0, IMAP_BUFSIZE);

  return c;
}

static int imap_getline(ImapContext *c) {
  if(!fgets(c->buf, IMAP_BUFSIZE, c->in)) {
    if(ferror(c->in))
      imap_error(c, "read: %s", strerror(errno));
    else
      imap_error(c, "read: unexpected eof");
    imap_close(c);
    return -1;
  }
  return 0;
}

static int imap_command(ImapContext *c, UntaggedCallback ucb, 
    const char *fmt, ...) {
  va_list ap;

  fputs(". ", c->out);
  va_start(ap, fmt);
  vfprintf(c->out, fmt, ap);
  va_end(ap);
  fputs("\r\n", c->out);
  fflush(c->out);

  while(1) {
    char *resp[3];
    int nresp;

    if(imap_getline(c) < 0)
      return -1;

    util_chomp(c->buf);
    nresp = util_split(" ", c->buf, resp, 3);
    if(nresp < 2) {
      imap_close(c);
      imap_error(c, "Confusing imap response: %s %s %s", nresp ? resp[0] : "",
          nresp > 1 ? resp[1] : "", nresp > 2 ? resp[2] : "");
      return -1;
    }
    if(!strcmp(resp[0], ".")) {
      if(!strcasecmp(resp[1], "OK"))
        return 1;
      else if(!strcasecmp(resp[1], "BAD"))
        return 0;
      else if(!strcasecmp(resp[1], "NO"))
        return 0;
      else {
        imap_close(c);
        imap_error(c, "Confusing imap response: %s %s %s", nresp ? resp[0] : "",
            nresp > 1 ? resp[1] : "", nresp > 2 ? resp[2] : "");
        return -1;
      }
    }
    else if(!strcmp(resp[0], "*")) 
      ucb(c, resp[1], resp[2]);
    else {
      imap_close(c);
      imap_error(c, "Confusing imap response: %s %s %s", nresp ? resp[0] : "",
          nresp > 1 ? resp[1] : "", nresp > 2 ? resp[2] : "");
      return -1;
    }
  }
}

ImapContext *imap_preauth(char * const cmd[]) {
  ImapContext *c;
  pid_t pid;
  int read, write;
  char *resp[3];
  int nresp;

  c = imap_new();
  if(!c)
    return NULL;

  pid = util_pipeto(cmd, &read, &write);
  if(!pid) {
    imap_error(c, "%s", strerror(errno));
    return c;
  }

  c->in = fdopen(read, "r");
  if(!c->in) {
    close(read);
    close(write);
    imap_error(c, "%s", strerror(errno));
    return c;
  }
  c->out = fdopen(write, "w");
  if(!c->out) {
    fclose(c->in);
    close(write);
    imap_error(c, "%s", strerror(errno));
    return c;
  }

  if(imap_getline(c) < 0)
    return c;
  util_chomp(c->buf);
  nresp = util_split(" ", c->buf, resp, 3);
  if(nresp < 2 || strcmp(resp[0], "*") || strcasecmp(resp[1], "PREAUTH")) {
    imap_close(c);
    imap_error(c, "Unexpected preauth greeting: %s %s %s", 
        nresp ? resp[0] : "", nresp > 1 ? resp[1] : "", 
        nresp > 2 ? resp[2] : "");
    return c;
  }

  return c;
}

ImapContext *imap_login(const char *host, const char *login, const char *pass) {
  ImapContext *c;
  char *p;
  unsigned short port;
  int fd;
  char *resp[3];
  int nresp;

  c = imap_new();
  if(!c)
    return NULL;

  p = strchr(host, ':');
  if(!p)
    port = IMAP_DEFAULT_PORT;
  else {
    *p++ = '\0';
    port = atoi(p);
  }

  fd = util_iconnect(host, port);
  if(fd < 0) {
    if(fd < -1) {
      imap_error(c, "Unable to resolve %s: %s", host, hstrerror(h_errno));
      return c;
    }
    else {
      imap_error(c, "Unable to connect to %s/%hu: %s", host, port, 
          strerror(errno));
      return c;
    }
  }

  c->in = fdopen(fd, "r");
  if(!c->in) {
    close(fd);
    imap_error(c, "%s", strerror(errno));
    return c;
  }
  c->out = fdopen(fd, "w");
  if(!c->out) {
    fclose(c->in);
    imap_error(c, "%s", strerror(errno));
    return c;
  }

  if(imap_getline(c) < 0)
    return c;
  util_chomp(c->buf);
  nresp = util_split(" ", c->buf, resp, 3);
  if(nresp < 2 || strcmp(resp[0], "*") || strcasecmp(resp[1], "OK")) {
    imap_close(c);
    imap_error(c, "Unexpected IMAP greeting: %s %s %s", 
        nresp ? resp[0] : "", nresp > 1 ? resp[1] : "", 
        nresp > 2 ? resp[2] : "");
    return c;
  }

  switch(imap_command(c, handle_ignore, "LOGIN %s %s", login, pass)) {
    case -1:
      return c;
    case 0:
      imap_close(c);
      imap_error(c, "LOGIN command failed");
      return c;
    default:
      return c;
  }
}

void imap_destroy(ImapContext *c) {
  imap_close(c);
  free(c);
}

int imap_get_recent(ImapContext *c, const char *folder) {
  switch(imap_command(c, handle_status, "STATUS %s (RECENT)", folder)) {
    case -1:
      return -1;
    case 0:
      return 0;
    default:
      return c->recent;
  }
}

